#include "M5Atom.h"
#include <WiFi.h>
#include "icon.h"
#include <vector>
#include <Preferences.h>

#define USE_STATIC_IP
#define ID 1
#define ATOMICON_PRESSED_COLOR  0xFF, 0xAA, 0x00

#ifdef USE_STATIC_IP
IPAddress LOCAL_IP(192, 168, 4, 100 + ID);
IPAddress GATEWAY(192, 168, 4, 1);
IPAddress SUBNET(255, 255, 0, 0);
IPAddress PRIMARY_DNS(8, 8, 8, 8); //optional
IPAddress SECOND_DNS(8, 8, 4, 4);  //optional
#endif

typedef std::vector<String> string_vec_t;

const char *PASSWORD = "12345678";
#define UDP_PORT 1001
WiFiUDP udp;
Preferences prefs;

typedef struct
{
    uint8_t bitmap[25*3];
    String func;
}atomkey_t;

std::vector<atomkey_t> key_vec;
int8_t key_state = 1;

const unsigned char icon_err[25*3] = {
    255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 0, 0,
    0, 0, 0, 255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0,
    255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 0, 0,
};

std::vector<String> split(String src, String delim)
{
	std::vector<String> buf;
    int32_t flag = 0, pst = 0;
    int32_t len = delim.length();

    while (1)
    {
        flag = src.indexOf(delim, flag);
        if (flag < 0)
            break;
        buf.push_back(src.substring(pst, flag));
        pst = flag + len;
        flag += len;
    }

    if (pst < src.length())
    {
        buf.push_back(src.substring(pst, src.length()));
    }

    return buf;
}

void updateIcon(uint8_t *icon, uint8_t c1, uint8_t c2, uint8_t c3)
{
    uint8_t buf[25*3+2] = {5, 5};
    for(int i = 0; i < 25*3; i+=3)
    {
        if(icon[i] || icon[i+1] || icon[i+2])
        {
            buf[2+i] = c1;
            buf[2+i+1] = c2;
            buf[2+i+2] = c3;
        }
    }
    M5.dis.displaybuff(buf);
}

void updateIcon(uint8_t *icon)
{
    uint8_t buf[25*3+2 + 255] = {5, 5};
    memcpy(buf + 2, icon, 25*3);
    M5.dis.displaybuff(buf);
}

void saveConfig(std::vector<atomkey_t> &keys)
{
    char buf[20];
    prefs.begin("config");
    prefs.putUChar("total", keys.size());
    for(int i = 0; i < keys.size(); i++)
    {
        sprintf(buf, "func%d", i);
        prefs.putString(buf, keys.at(i).func);
        sprintf(buf, "bitmap%d", i);
        prefs.putBytes(buf, keys.at(i).bitmap, 25*3);
    }
}

std::vector<atomkey_t> loadConfig()
{
    std::vector<atomkey_t> keys;
    atomkey_t key;
    char buf[20];

    prefs.begin("config");
    uint8_t total = prefs.getUChar("total", 0);
    if(total == 0)
    {
        key.func = "null";
        memcpy(key.bitmap, icon_err, 25*3);
        keys.push_back(key);
        return keys;
    }
    else
    {
        for(int i = 0; i < total; i++)
        {
            sprintf(buf, "func%d", i);
            key.func = prefs.getString(buf);
            sprintf(buf, "bitmap%d", i);
            prefs.getBytes(buf, key.bitmap, 25*3);
            keys.push_back(key);
        }
        return keys;
    }
}

void udpRecive()
{
    uint32_t udplen = udp.parsePacket();
    if(udplen)
    {
        uint8_t data[udplen];
		udp.read(data, udplen);
        data[udplen - 1] = '\0';
        String payload((char*)data);
        Serial.println(payload);

        std::vector<String> r = split(payload, "|");

        if(r.at(0) == "Setting")
        {
            int id = r.at(3).toInt();
            if(id == ID)
            {
                if(r.at(1) == "Remove")
                {
                    int index = r.at(2).toInt();
                    if(index > 0 && index < key_vec.size())
                    {
                        key_vec.erase(std::begin(key_vec) + index);
                        saveConfig(key_vec);
                        key_state = 1;
                        updateIcon(key_vec.at(key_state).bitmap);
                    }
                    else if(index == 0)
                    {
                        key_vec.clear();
                        prefs.begin("config");
                        prefs.clear();
                        prefs.putUChar("total", key_vec.size());
                        key_vec = loadConfig();
                        key_state = 0;
                        updateIcon(key_vec.at(key_state).bitmap);
                    }
                    return;
                }

                atomkey_t key;
                std::vector<String> c = split(r.at(4), ",");
                if(c.size() != 25*3)
                {
                    Serial.printf("Error|Invalid Color buffer\n");
                    return;
                }
                else
                {
                    for(int i = 0; i < c.size(); i++)
                    {
                        key.bitmap[i] = c.at(i).toInt();
                    }
                }
                key.func = "Action|" + r.at(1) + "|" + r.at(2) + "\n";
                int index = r.at(5).toInt();
                if(index == 0 || index >= key_vec.size())
                {
                    key_vec.push_back(key);
                }
                else
                {
                    key_vec.at(index) = key;
                }
                saveConfig(key_vec);

                if(key_state == 0)
                {
                    key_vec = loadConfig();
                    key_state = 1;
                    updateIcon(key_vec.at(key_state).bitmap);
                }
            }
        }
    }
}

void udpSendString(String str)
{
    udp.beginPacket(IPAddress(192, 168, 4, 1), 1001);
    udp.write((uint8_t *)str.c_str(), str.length());
    udp.endPacket();
}

void setup()
{
    M5.begin(true, false, true);
    M5.dis.setBrightness(10);
    delay(50);
    WiFi.mode(WIFI_STA);
    Serial.printf("Atom Board Slave %d started.\n", ID);

    delay(200 + ID*100);
#ifdef USE_STATIC_IP
    if (!WiFi.config(LOCAL_IP, GATEWAY, SUBNET, PRIMARY_DNS, SECOND_DNS))
    {
        ESP.restart();
    }
#endif
    WiFi.begin("AtomBoard_Host", PASSWORD);
    while (WiFi.status() != WL_CONNECTED)
    {
        delay(500);
        Serial.print(".");
    }
    udp.begin(UDP_PORT);

    Serial.print("\nWiFi connected with IP: ");
    Serial.println(WiFi.localIP());

    key_vec = loadConfig();

    if(key_vec.size() == 1)
    {
        key_state = 0;
    }
    updateIcon(key_vec.at(key_state).bitmap);
}

void loop()
{
    udpRecive();

    if (M5.Btn.isPressed())
    {
        if(key_state > 0)
        {
            updateIcon(key_vec.at(key_state).bitmap, ATOMICON_PRESSED_COLOR);
            udpSendString(key_vec.at(key_state).func);
            key_state = -key_state;
        }
    }
    else if(key_state < 0)
    {
        key_state = -key_state;
        key_state ++;
        if(key_state >= key_vec.size())
        {
            key_state = 1;
        }
        updateIcon(key_vec.at(key_state).bitmap);
    }

    M5.update();
}